
// Standardne pre MS win aplikacie
#include <windows.h>
#include <windowsx.h>

// GLEW kniznica. pre spravu novsich rozsireni
#include "glew.h"
#include "wglew.h"

#include <stdio.h>
#include <math.h>

#include "shaders.h"
#include "textures.h"
#include "geometry.h"
#include "fonts.h"

// Globalne premenne
GLboolean g_isAccumAA = false;
GLboolean g_isMultisampling = true;
GLboolean g_isTexture = false;
float sun_rotation = 0.0f;
bool g_finished = false;

// Win32 parameter okna
HWND g_hwnd;
HDC g_hdc;
HGLRC g_hrc;
GLuint g_window_height, g_window_width;

// data pre AA s pouzitim accumulation buffra
typedef struct {	
   GLfloat x, y;
} jitter_point;
jitter_point j8[] ={ {-0.334818,  0.435331}, {0.286438, -0.393495}, {0.459462,  0.141540}, {-0.414498, -0.192829}, {-0.183790,  0.082102}, {-0.079263, -0.317383}, {0.102254,  0.299133}, {0.164216, -0.054399}};


// Globalne premenne pre prezeranie sceny
GLdouble vzd, old_vzd, fi, xi, old_fi, old_xi;
GLint left_mouse_down_x, left_mouse_down_y;
GLint right_mouse_down_y;
GLboolean right_down = false, left_down = false;

// id textur
GLuint tex_sun, tex_earth, tex_lava, tex_phobos, tex_skybox;

// id shaderov
GLuint shaders_skybox, shaders_perpixel_phong, shaders_envmap;


void initShaders()
{
    // nacitanie, kompilacia a zlinkovanie shaderov
    shaders_skybox = loadProgram("shaders/skybox.vert", "shaders/skybox.frag");
    SetShaderUniform1i(shaders_skybox, "cube_map", 0);

    shaders_perpixel_phong = loadProgram("shaders/perpixel_phong.vert", "shaders/perpixel_phong.frag");
    SetShaderUniform1i(shaders_perpixel_phong, "color_texture", 0);

    shaders_envmap = loadProgram("shaders/perpixel_envmap.vert", "shaders/perpixel_envmap.frag");
    SetShaderUniform1i(shaders_envmap, "color_texture", 0);
    SetShaderUniform1i(shaders_envmap, "cubemap_texture", 1);

    glActiveTexture(GL_TEXTURE0);
}

// inicializacia OpenGL textur
void initTextures()
{
    // ************* prva textura, zem *************
    tex_earth = LoadTexture("textures/earth.jpg");

    // ************* druha textura, slnko *************
    tex_sun = LoadTexture("textures/sun.jpg");

    // ************* tretia textura, jupiter *************
    tex_lava = LoadTexture("textures/lava.png");

    // ************* stvrta textura, phobos *************
    tex_phobos = LoadTexture("textures/phobos.jpg");

    // ************* stvrta textura, phobos *************
    tex_skybox = LoadCubeTexture("skybox/snow_positive_x.jpg", "skybox/snow_negative_x.jpg",
                                 "skybox/snow_positive_y.jpg", "skybox/snow_negative_y.jpg",
                                 "skybox/snow_positive_z.jpg", "skybox/snow_negative_z.jpg");
}

// Open GL Inicializacia
bool init(void)
{
    // Cierne pozadie
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    // Nastav z-buffer testovanie
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // nastav stencil buffer
    glClearStencil(0);

    // nastav zmiesavaciu funkciu
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // nastav vhodne mapovanie textur
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    // initializuj kameru
    fi = 0.7f; xi = 0.7f; vzd = 10.0f;

    // initializuje svetlo
    GLfloat light_ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    GLfloat light_diffuse[] = { 0.9f, 0.9f, 0.9f, 1.0f };
    GLfloat light_specular[] = { 0.5f, 0.5f, 0.5f, 1.0f };
    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
    glEnable(GL_LIGHT0);

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    // parametre hmly
    glFogi(GL_FOG_MODE, GL_LINEAR);
    GLfloat fogColor[4]= {0.5f, 0.5f, 0.5f, 1.0f};
    glFogfv(GL_FOG_COLOR, fogColor);
    glFogf(GL_FOG_DENSITY, 0.35f);
    glHint(GL_FOG_HINT, GL_DONT_CARE);
    glFogf(GL_FOG_START, 3.0f);
    glFogf(GL_FOG_END, 20.0f);
    glEnable(GL_FOG);

	return true;
}

void renderScene()
{
    // Vymaz (vypln) obrazovku a z-buffer definovanymi hodnotami
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	

    // Nastav modelview maticu (transformaciu) na jednotkovu maticu
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();									

    // Nastav view matice pre pohlad na svetovy suradnicovy system
    GLfloat camera_pos[4];
    camera_pos[0] = vzd*cos(fi)*cos(xi);
	camera_pos[1] = vzd*sin(fi)*cos(xi);
	camera_pos[2] = vzd*sin(xi);
    camera_pos[3] = 1;
	gluLookAt(camera_pos[0], camera_pos[1], camera_pos[2], 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f );
    float view_matrix[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, view_matrix);
    SetShaderUniformMatrix4f(shaders_envmap, "view_matrix", view_matrix);

    // vykresli suradnicove osi svetoveho systemu roznymi farbami
    glLineWidth(1);
    glBegin( GL_LINES );
    glColor3f(1.0f, 0.0f, 0.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 100.0f, 0.0f, 0.0f );
    glColor3f(0.0f, 1.0f, 0.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 0.0f, 100.0f, 0.0f );
    glColor3f(0.0f, 0.0f, 1.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 0.0f, 0.0f, 100.0f );
    glEnd();

    // vykresli skybox
    // skybox je dany vo svetovych suradniciach, akurat jeho stred je stale v bode, kdeje kamera
    glUseProgram(shaders_skybox);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_CUBE_MAP, tex_skybox);
    glPushMatrix();
    glScalef(500, 500, 500);
    renderCubeVBO();
    glPopMatrix();
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glUseProgram(0);

    // using per-pixel phong
    glUseProgram(shaders_perpixel_phong);
    if (g_isTexture)
        SetShaderUniform1i(shaders_perpixel_phong, "texturing_enabled", 1);
    else
        SetShaderUniform1i(shaders_perpixel_phong, "texturing_enabled", 0);

    // vykresli slnko
    SetShaderUniform1i(shaders_perpixel_phong, "lighting_enabled", 0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex_sun);
    glPushMatrix();
    glRotatef(sun_rotation, 0.0f, 0.0f, 1.0f);
    glTranslatef(6.0f, 0.0f, 0.0f);
    // nastav poziciu svetla do stredu slnka
    GLfloat sun_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    glLightfv(GL_LIGHT0, GL_POSITION, sun_position);
    glScalef(0.3f, 0.3f, 0.3f);
    glColor3f(1.0f, 1.0f, 1.0f );
    renderSphereVBO();
    glPopMatrix();

    // vykresli phobos
    SetShaderUniform1i(shaders_perpixel_phong, "lighting_enabled", 1);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex_phobos);
    glPushMatrix();
    glTranslatef(0.0f, 0.0f, 1.0f);
    glScalef(1.0f, 1.0f, 1.0f);
    renderSphereVBO();
    glPopMatrix();
    
    // using per-pixel environment mapping
    glUseProgram(shaders_envmap);
    if (g_isTexture)
        SetShaderUniform1i(shaders_envmap, "texturing_enabled", 1);
    else
        SetShaderUniform1i(shaders_envmap, "texturing_enabled", 0);

    // vykresli 2x torus s environment mapou
    glPushMatrix();
    glLoadIdentity();
    SetShaderUniformMatrix4f(shaders_envmap, "view_matrix", view_matrix);
    SetShaderUniform1i(shaders_envmap, "color_texture", 0);
    SetShaderUniform1i(shaders_envmap, "cubemap_texture", 1);
    SetShaderUniform4f(shaders_envmap, "camera_world", camera_pos);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex_lava);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_CUBE_MAP, tex_skybox);
    glTranslatef(3, 0, 1);
    renderTorusVBO();
    glTranslatef(-6, 0, 0);
    renderTorusVBO();
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glActiveTexture(GL_TEXTURE0);
    glPopMatrix();
    glUseProgram(0);

    // vykresli 2x stvorec v jednej rovine
    glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
    glBegin(GL_QUADS);
    glVertex3f(-1.0f, -4.0f, 0.0f);
    glVertex3f(4.0f, -4.0f, 0.0f);
    glVertex3f(4.0f, 4.0f, 0.0f);
    glVertex3f(-1.0f, 4.0f, 0.0f);
    glEnd();
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.0, 1.0);
    glColor4f(0.0f, 0.0f, 1.0f, 1.0f);
    glBegin(GL_QUADS);
    glVertex3f(1.0f, -4.0f, 0.0f);
    glVertex3f(-4.0f, -4.0f, 0.0f);
    glVertex3f(-4.0f, 4.0f, 0.0f);
    glVertex3f(1.0f, 4.0f, 0.0f);
    glEnd();
    glDisable(GL_POLYGON_OFFSET_FILL);

    // vykresli 3D text
    glEnable(GL_LIGHTING);
    glColor3f(0.0f, 1.0f, 1.0f);
    render3DTextWGL(7, 0, 3, 1, "Environment mapping & multitexturing");
    glDisable(GL_LIGHTING);

    // vykresli 2D text
    glColor3f(1.0f, 1.0f, 0.0f);
    render2DTextFT(10, g_window_height - 20, "Projekt OpenGL - Antialiasing");
    render2DTextFT(10, g_window_height - 45, "F1 - zapni/vypni Accum AA");
    render2DTextFT(10, g_window_height - 70, "F2 - zapni/vypni textury");
    render2DTextFT(10, g_window_height - 95, "F2 - zapni/vypni Multisample AA");
}
// Procedura obsluhujuca vykreslovanie
void render(void)   
{
    if (g_isAccumAA)
    {
        // viacnasobne renderovanie sceny do akumulacneho bufra
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT);

        // vypocitaj parametre pohladoveho objemu
        GLint viewport[4];
        glGetIntegerv(GL_VIEWPORT, viewport);
        const double PI = 3.14159265358979323846;
	    GLdouble fov2 = ((60*PI) / 180.0) / 2.0;
        GLdouble cam_near = 0.1;
        GLdouble cam_far = 1000;
	    GLdouble cam_top = cam_near / (cos(fov2) / sin(fov2));
	    GLdouble cam_bottom = -cam_top;
	    GLdouble cam_right = cam_top * viewport[2] / float(viewport[3]);
	    GLdouble cam_left = -cam_right;

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glMatrixMode(GL_MODELVIEW);

        int numSamples = 8;
        // pre niekolko roznych nastaveni kamery, renderuj scenu
        for (int jitter = 0; jitter < numSamples; jitter++) 
        {
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glMatrixMode(GL_PROJECTION);
            glLoadIdentity();
            GLdouble dx = -j8[jitter].x * (cam_right - cam_left) / (GLdouble) viewport[2];
            GLdouble dy = -j8[jitter].y * (cam_top - cam_bottom) / (GLdouble) viewport[3];
            glFrustum(cam_left + dx, cam_right + dx, cam_bottom + dy, cam_top + dy, cam_near, cam_far);
            glMatrixMode(GL_MODELVIEW);

            renderScene();

            glAccum(GL_ACCUM, 1.0/float(numSamples));
        }
        glAccum(GL_RETURN, 1.0);

        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
    }
    else
    {
        renderScene();
    }
}

// Procedura obsluhujuca zmenu rozmerov okna
void reshape(int w, int h)
{
    g_window_width = w;
    g_window_height = h;

    // nastvenie casti okna pre vykreslenie, tentoraz kreslime do celeho okna
	glViewport(0, 0, w, h);
    // Budeme menit projekcnu maticu (transformaciu)
	glMatrixMode(GL_PROJECTION);                        
	glLoadIdentity();                                  

	if (h == 0) 
        h = 1;
    // Chceme perspektivnu projekciu
    float ratio = (float)w/(float)h;
	gluPerspective(60, ratio, 0.1, 1000.0);
    //glOrtho(-10.0f*ratio, 10.0f*ratio, -10.0f, 10.0f, 1.0, 5000.0); 
}

// Procedura obsluhujuca stlacenie klavesy
void keyboard(int key, int x, int y)
{
    switch (key) 
    {
        case VK_ESCAPE:
            // Ukoncenie programu
			g_finished = true;
		    break;
        // Ak F1 bolo stlacene
		case VK_F1:
            g_isAccumAA = !g_isAccumAA;
            RedrawWindow(g_hwnd, NULL, NULL, RDW_INVALIDATE);
		    break;
        // Ak F2 bolo stlacene
        case VK_F2:
            g_isTexture = !g_isTexture;
            RedrawWindow(g_hwnd, NULL, NULL, RDW_INVALIDATE);
		    break;
        // Ak F3 bolo stlacene
        case VK_F3:
            g_isMultisampling = !g_isMultisampling;
            if (g_isMultisampling)
                glEnable(GL_MULTISAMPLE);
            else
                glDisable(GL_MULTISAMPLE);
            RedrawWindow(g_hwnd, NULL, NULL, RDW_INVALIDATE);
		    break;
		default:
		    break;
	}
}

// Procedura obsluhujuca stlacenie tlacitka mysi
void mouse_klik(int button, int state, int x, int y )
{
	switch (button) 
    {
		case 0:
            if (state == 0)
            {
                left_down = true;
			    left_mouse_down_x = x;
			    left_mouse_down_y = y;
			    old_fi = fi;
			    old_xi = xi;
                break;
            }
            else if (state == 1)
            {
                left_down = false;
            }
		    break;
        case 1:
            if (state == 0)
            {
                right_down = true;
			    right_mouse_down_y = y;
			    old_vzd = vzd;
                break;
            }
            else if (state == 1)
            {
                right_down = false;
            }
		    break;
		default:
		    break;
	}

    RedrawWindow(g_hwnd, NULL, NULL, RDW_INVALIDATE);
}

// Procedura obsluhujuca pohyb mysi
void mouse_move(int x, int y)
{
    if (left_down == GL_TRUE)
    {
        fi = old_fi + (left_mouse_down_x - x)/200.0f;
        xi = old_xi + (y - left_mouse_down_y)/200.0f;
        RedrawWindow(g_hwnd, NULL, NULL, RDW_INVALIDATE);
    }

    if (right_down == GL_TRUE)
    {
        vzd = old_vzd + (right_mouse_down_y - y)/10.0f;
        RedrawWindow(g_hwnd, NULL, NULL, RDW_INVALIDATE);
    } 

}

// funkcia pre spracovanie sprav, ktore prisli zo systemu pre nase okno
LRESULT CALLBACK OpenGLWindowProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{	

		case WM_CREATE:
		{
		}
		return 0;

        case WM_PAINT:
		{
            render();
            // Zamen buffre a zobraz vykreslene objekty
            SwapBuffers(g_hdc);
		}
		return 0;

		case WM_CLOSE:
			g_finished = true;
		return 0;

		case WM_SIZE:
			switch (wParam)
			{
				case SIZE_MINIMIZED:
				return 0;

				case SIZE_MAXIMIZED:
					reshape(LOWORD (lParam), HIWORD (lParam));
				return 0;

				case SIZE_RESTORED:
					reshape(LOWORD (lParam), HIWORD (lParam));
				return 0;
			}
		break;

		case WM_KEYDOWN:
			if ((wParam >= 0) && (wParam <= 255))
			{
                POINT p;
                GetCursorPos(&p);
                ScreenToClient(hWnd, &p);
				keyboard((int)wParam, p.x, p.y);
                return 0;
			}
		break;

		case WM_KEYUP:
			if ((wParam >= 0) && (wParam <= 255))
			{
				POINT p;
                GetCursorPos(&p);
                ScreenToClient(hWnd, &p);
				//keyboard((int)wParam, p.x, p.y);
				return 0;
			}
		break;

        case WM_LBUTTONDOWN:
				mouse_klik(0, 0, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
				return 0;

        case WM_LBUTTONUP:
				mouse_klik(0, 1, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
				return 0;

        case WM_RBUTTONDOWN:
				mouse_klik(1, 0, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
				return 0;

        case WM_RBUTTONUP:
				mouse_klik(1, 1, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
				return 0;

        case WM_MOUSEMOVE:
				mouse_move(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
				return 0;

	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);					// Pass Unhandled Messages To DefWindowProc
}

bool CreateWindowGL(int width, int height, bool fullscreen, bool multisample)
{
    printf("Creating OpenGL window...\n");

    // Registracia triedy pre nase OpenGL okno
	WNDCLASSEX windowClass;
	ZeroMemory (&windowClass, sizeof(WNDCLASSEX));
	windowClass.cbSize = sizeof (WNDCLASSEX);
	windowClass.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	windowClass.lpfnWndProc = (WNDPROC)(OpenGLWindowProc);
    windowClass.hInstance = GetModuleHandle(NULL);
	windowClass.hbrBackground = (HBRUSH)(COLOR_APPWORKSPACE);
	windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);
	windowClass.lpszClassName = "OpenGLWinClass";
	if (RegisterClassEx(&windowClass) == 0)
	{
		printf("Registration of OpenGL window class failed.\n");
		return false;
	}

    // ak chceme fullscreen, tak okno bude mat velkost obrazovky a bude bez okrajov a list
    DWORD windowStyle = WS_OVERLAPPEDWINDOW;
	DWORD windowExtendedStyle = WS_EX_APPWINDOW;
	if (fullscreen == true)
	{
        // zmen rozmery a styl okna
        width = GetSystemMetrics(SM_CXSCREEN);
        height = GetSystemMetrics(SM_CYSCREEN);
	    windowStyle = WS_POPUP;
		windowExtendedStyle |= WS_EX_TOPMOST;
        printf("Fullscreen mode...\n");
	}
	else
	{
		printf("Windowed mode...\n");
	}

    // vypocitaj skutocne rozmery okna na zaklade rozmerov klientskej casti okna
    RECT windowRect = {0, 0, width, height};
    AdjustWindowRectEx (&windowRect, windowStyle, 0, windowExtendedStyle);

	// vytvor okno pomocou Win32 API
	HWND g_hwnd = CreateWindowEx (windowExtendedStyle,
								   "OpenGLWinClass",
								   "OpenGL window",
								   windowStyle,
								   0, 0,
								   windowRect.right - windowRect.left,
								   windowRect.bottom - windowRect.top,
								   HWND_DESKTOP,
								   0,
								   GetModuleHandle(NULL),
								   NULL);

    // otestuj ci sa okno vytvorilo
	if (g_hwnd == 0)
	{
        printf("ERROR: Failed to create OpenGL window.\n");
		return false;
	}

    // ziskaj device context okna
	g_hdc = GetDC(g_hwnd);
	if (g_hdc == 0)
	{
		printf("ERROR: Can not retrieve device context for OpenGL window.\n");
		DestroyWindow(g_hwnd);
		return false;
	}

    // struktura v ktorej su vlastnosti kazdeho pixelu pre nase nove okno
	PIXELFORMATDESCRIPTOR pfd =
	{
		sizeof (PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		24,           // presnost color buffera v bitoch  
		0, 0, 0, 0, 0, 0,
		1,
		0,
		24,           // presnost accumulation buffera v bitoch
		0, 0, 0, 0,
		16,           // presnost depth buffera v bitoch
		0,
		0,
		PFD_MAIN_PLANE,
		0,
		0, 0, 0
	};

    // ziskaj pixel format ktory sa co najviac podoba danemu popisu
	GLuint PixelFormat = ChoosePixelFormat(g_hdc, &pfd);
	if (PixelFormat == 0)
	{
        printf("ERROR: Cannot find pixel format for OpenGL window.\n");
		ReleaseDC(g_hwnd, g_hdc);
		DestroyWindow(g_hwnd);
		return false;
	}

    // nastav pixel format pre dane okno
	if (SetPixelFormat(g_hdc, PixelFormat, &pfd) == FALSE)		// Try To Set The Pixel Format
	{
		printf("ERROR: Cannot set pixel format for OpenGL window.\n");
		ReleaseDC(g_hwnd, g_hdc);
		DestroyWindow(g_hwnd);
		return false;
	}

    // vytvor OpenGL kontext pre okno s danym pixel formatom
	g_hrc = wglCreateContext(g_hdc);
	if (g_hrc == 0)
	{
		printf("ERROR: Cannot create OpenGL rendering context.\n");
		ReleaseDC(g_hwnd, g_hdc);
		DestroyWindow(g_hwnd);
		return false;
	}

	// nastva rendering kontext ako aktivny
	if (wglMakeCurrent(g_hdc, g_hrc) == false)
	{
		printf("ERROR: Cannot set OpenGL rendering context as current.\n");
		wglDeleteContext(g_hrc);
		ReleaseDC(g_hwnd, g_hdc);
		DestroyWindow(g_hwnd);
		return false;
	}

	// mame fungujuci OpenGL kontext, mozme zistit, ci exsituje pixel format s multisamplingom
    if (multisample)
    {
        // nastav atributy
        int attributes[] = 
        {
            WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
            WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
            WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
            WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_ARB,
            WGL_COLOR_BITS_ARB, pfd.cColorBits,
            WGL_DEPTH_BITS_ARB, pfd.cDepthBits,
            WGL_STENCIL_BITS_ARB, pfd.cStencilBits,
            WGL_ACCUM_BITS_ARB, pfd.cAccumBits,
            WGL_SAMPLE_BUFFERS_ARB, GL_TRUE,              // zapni multisampling
            WGL_SAMPLES_ARB, 4,                     // pocet samplov pre multisampling
            0
        };

        glewInit();

        // zisti, ci dany pixel format sa v systeme nachadza
        int pixelFormatMultisample;
        UINT numFormats;
        float fAttributes[] = {0,0};
        BOOL valid = wglChoosePixelFormatARB(g_hdc, attributes, fAttributes, 1, &pixelFormatMultisample, &numFormats);
        if (valid && numFormats >= 1)
        {
            // pixel format sa nasiel, preto odstran stare okno, stary kontext a vytvor novy s multisamplingom
            wglMakeCurrent(g_hdc, NULL);
	        ReleaseDC(g_hwnd, g_hdc);
	        DestroyWindow(g_hwnd);

            g_hwnd = CreateWindowEx (windowExtendedStyle,
								   "OpenGLWinClass",
								   "OpenGL window",
								   windowStyle,
								   0, 0,
								   windowRect.right - windowRect.left,
								   windowRect.bottom - windowRect.top,
								   HWND_DESKTOP,
								   0,
								   GetModuleHandle(NULL),
								   NULL);
            g_hdc = GetDC(g_hwnd);
            SetPixelFormat(g_hdc, pixelFormatMultisample, &pfd); 
            g_hrc = wglCreateContext(g_hdc);
            if (g_hrc == 0)
	        {
		        printf("ERROR: Cannot create multisample OpenGL rendering context.\n");
		        ReleaseDC(g_hwnd, g_hdc);
		        DestroyWindow(g_hwnd);
		        return false;
	        }

            if (wglMakeCurrent(g_hdc, g_hrc) == false)
	        {
		        printf("ERROR: Cannot set multisample OpenGL rendering context as current.\n");
		        wglDeleteContext(g_hrc);
		        ReleaseDC(g_hwnd, g_hdc);
		        DestroyWindow(g_hwnd);
		        return false;
	        }

            // zapni multisampling
            if (g_isMultisampling)
                glEnable(GL_MULTISAMPLE);

            GLint samples;
            glGetIntegerv(GL_SAMPLES, &samples);
            printf("Using context with %ix multisampling.\n", samples);
        }

    }
    else
    {
        printf("Using context with no multisampling.\n");
    }

    // zobraz vytvorene OpenGL okno
	ShowWindow(g_hwnd, SW_NORMAL);
	reshape(width, height);

    printf("OpenGL window successfully created.\n");

	return true;
}

bool DestroyWindowGL()
{
	wglMakeCurrent(g_hdc, NULL);
	ReleaseDC(g_hwnd, g_hdc);
	DestroyWindow(g_hwnd);

	return true;
}

// Zakladna funkcia spustena na zaciatku
int main(int argc, char** argv)
{
    if (!CreateWindowGL(800, 600, false, true))
        return 0;

    // Zinicializujeme GLEW
    glewInit();

    // inicializacia
    init();
    initShaders();
    initTextures();
    initSphereVBO(32, 32);
    initTorusVBO(0.6f, 2.0f, 32, 32);
    init2DFontFT("fonts/optien.ttf");

    // Spustenie hlavneho okruhu zachytavania sprav
	while (!g_finished)
    {
        MSG msg;	
        if (PeekMessage(&msg, g_hwnd, 0, 0, PM_REMOVE) != 0)
        {
            if (msg.message != WM_QUIT)
            {
                DispatchMessage(&msg);
            }
            else
            {
                g_finished = true;
            }
        }
    }

    DeleteAllVBO();
    DestroyWindowGL();

	return 0;
}
